/***************************************************************************
***
***    Copyright 2007  Hon Hai Precision Ind. Co. Ltd.
***    All Rights Reserved.
***    No portions of this material shall be reproduced in any form without the
***    written permission of Hon Hai Precision Ind. Co. Ltd.
***
***    All information contained in this document is Hon Hai Precision Ind.  
***    Co. Ltd. company private, proprietary, and trade secret property and 
***    are protected by international intellectual property laws and treaties.
***
****************************************************************************/

/*******************************************************************************
 ***  Filename:
 ***      wps_monitor.c
 ***
 ***  Description:
 ***      -- Monitor WPS push button status
 ***      -- Control WPS LED status
 ***
 ***  History:
 ***
 ***   Modify Reason                    Author          Date         Search Flag(Option)
 ***--------------------------------------------------------------------------------------
 ***   Initial version                  pling           07/16/007 
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <time.h>

#include <epivers.h>
#include <shutils.h>
#include <bcmdevs.h>
#include <bcmparams.h>
#include <bcmgpio.h>
#include <bcmnvram.h>

#include "wps_led.h"
#include "wps_monitor.h"

#define DEV_WPS_LED             "/dev/wps_led"

#define ACTION_NONE             0
#define ACTION_BLINK_TOGGLE     1
#define ACTION_BLINK_WPS        2

#define LOCAL       static


static int check_wps_button(int gpio_status)
{
    unsigned long value;
    int i;

    value = gpio_status & GPIO_WPS_BUTTON_MASK;
    if (value)
        return ACTION_NONE;

    /* Wait up to 3 second for user to release */
    for (i=0; i<30; i++)
    {
        bcmgpio_in(1 << GPIO_WPS_BUTTON, &value);

        /*
         * If button is released before the 3sec mark, 
         * toggle LED on / off
         */
        if (value)
            return ACTION_BLINK_TOGGLE;

        usleep(100000);
    }
    
    /* 3 sec passed, start WPS !!! */
    return ACTION_BLINK_WPS;
}

static int check_ant_status(int gpio_status)
{
    static int first_time = 1;
    int ant_status = gpio_status;

    ant_status >>= GPIO_11N_ANT1_TX;
    ant_status &= 0x07;

    /* If both 11N aren't active, we randomly set the ant status */
    if ((ant_status & 0x3) == 0)
    {
        int rand_num;

        if (first_time)
        {
            srand(time(NULL));
            first_time = 0;
        }
        rand_num = (rand() % 3) + 1;  /* 1 <= rand_num <= 3 */
        ant_status += rand_num;
    }

    return ant_status;
}

LOCAL int convert_ant_status_to_LED_pattern(int ant_status)
{
    /*
     * The following conversion of ANT status to LED patten follows
     * NETGEAR Patrick Rata's spec:
     *
     * 11n ant0 + 11n ant1  : blink LED 2, 4, 6, 8
     * 11n ant0 only        : blink LED 1, 4, 5, 8
     * 11n ant1 only        : blink LED 1, 2, 5, 6
     * 
     * 11g ant on           : blink LED 7
     * 11g ant off          : blink LED 3
     */

    /*
     * There is a output register to control the 8 LEDs:
     * bit 0 : LED 1, 5
     * bit 1 : LED 2, 6
     * bit 2 : LED 4, 8
     * bit 3 : LED 3
     * bit 4 : LED 7
     */
    
    int blink_pattern = 0;
    int ant0_11n, ant1_11n, ant_11g;

    ant0_11n = ant_status & 0x2;
    ant1_11n = ant_status & 0x1;
    ant_11g  = ant_status & 0x4;

    if (nvram_match("wl0_radio", "1"))
    {
        if (ant0_11n && ant1_11n)
            blink_pattern = 0x02 | 0x04;        /* bit 1 + bit 2 */
        else
        if (ant0_11n)
            blink_pattern = 0x01 | 0x04;        /* bit 0 + bit 2 */
        else
        if (ant1_11n)
            blink_pattern = 0x01 | 0x02;        /* bit 0 + bit 1 */
    }

    if (nvram_match("wl1_radio", "1"))
    {
        if (ant_11g)
            blink_pattern |= 0x10;              /* bit 4 */
        else
            blink_pattern |= 0x08;              /* bit 3 */
    }
    
    return blink_pattern;
}

LOCAL int send_wps_led_cmd(int cmd, int arg)
{
    int ret_val=0;
    int fd;

    fd = open(DEV_WPS_LED, O_RDWR);
    if (fd < 0) 
        return -1;

    ret_val = ioctl(fd, cmd, arg);
    close(fd);

    return ret_val;
}

int wps_check(int gpio_status)
{
    int status;
    static unsigned long loop_count = 0;
    static unsigned long wps_start_count = 0;
    static int wps_led_state = 1;
    static int ant_status;
    int wps_led_pattern;
    int wsc_proc_status;
    int abort_wps;

    /* 
     * If WPS is already started from GUI, 
     * then don't do anything here.
     */
    if (nvram_match("wps_start", "gui"))
        return 0;

    loop_count++;

    /* Check if WPS is start */
    if (wps_start_count != 0)
    {
        abort_wps = 0;

        /* Check whether WPS completes successfully or in error */
        wsc_proc_status = atoi(nvram_get("wsc_proc_status"));
        if (wsc_proc_status == 2 || wsc_proc_status == 3 || wsc_proc_status == 4)
        {
            printf("%s: wsc returns %d\n", __FUNCTION__, wsc_proc_status);
            abort_wps = 1;
        }

        /* Check whether WPS should time out now */
        if ((loop_count - wps_start_count) >= 480)
        {
            printf("%s: WPS time out\n", __FUNCTION__);
            abort_wps = 1;
        }

        if (abort_wps)
        {
            wps_start_count = 0;
 
            /* Send signal to wpsd to terminate WPS */
            system("killall -SIGUSR2 wpsd");

            /* Clear NVRAM flag */
            nvram_set("wps_start", "none");
            
            /* LEDs back to normal blink */
            if (wps_led_state)
                send_wps_led_cmd(LED_BLINK_NORMAL, 0);
            else
                send_wps_led_cmd(LED_BLINK_OFF, 0);

            /* Use previous ANT pattern */
            wps_led_pattern = convert_ant_status_to_LED_pattern(ant_status);
            send_wps_led_cmd(LED_BLINK_PATTERN, wps_led_pattern);
        }
    }
    else
    {
        /* Check WPS button status */
        status = check_wps_button(gpio_status);

        if (status == ACTION_BLINK_TOGGLE)
        {
           
            /* Don't start WPS / toggle LEDs while upgrading firmware */
            if (check_upgrade_status())
            {
                printf("Upgrading firmware, don't start WPS\n");
                return 0;
            }
           
            
            if (wps_led_state)
            {
                send_wps_led_cmd(LED_BLINK_OFF, 0);
                wps_led_state = 0;
            }
            else
            {
                send_wps_led_cmd(LED_BLINK_NORMAL, 0);
                wps_led_state = 1;
            }
        }
        else
        if (status  == ACTION_BLINK_WPS)
        {
          
            /* Don't start WPS / toggle LEDs while upgrading firmware */
            if (check_upgrade_status())
            {
                printf("Upgrading firmware, don't start WPS\n");
                return 0;
            }
          
            
            wps_start_count = loop_count;

            /* Tell WPS process that we use PBC method */
            nvram_set("wps_start", "pbc");
            nvram_set("wps_method", "pbc");

            /* Send signal to wpsd to start WPS */
            system("killall -SIGUSR1 wpsd");
        }

        
        /* Update LED blink pattern every 0.2 secs */
        //if (((loop_count % 2) == 0) && wps_start_count == 0)
        if (wps_start_count == 0)
        {
            ant_status = check_ant_status(gpio_status);
            wps_led_pattern = convert_ant_status_to_LED_pattern(ant_status);
            //printf("ant: %d, pattern: %x\n", ant_status, wps_led_pattern);
            send_wps_led_cmd(LED_BLINK_PATTERN, wps_led_pattern);
        }
    }

    return 0;
}
